"""Collection of experimental datapipes, explicitly not ready for production,
 but can be upgraded later, these might not be tested, and are to make quick experiments easier"""
from .ensure_n_nwp_variables import EnsureNNWPVariables
from .set_system_id_to_one import SetSystemIDsToOneIterDataPipe as SetSystemIDsToOne
